/*
 * backup.c
 *
 * Routines to perform that actual backup and restore of the device.
 *
 * Copyright (C) 2004 Daniel Thompson <rfbackup@redfelineninja.org.uk> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "rfbackup.h"

#include <stdarg.h>

#define MAX_DBNAME   32
#define MAX_FILENAME (MAX_DBNAME + 8)
#define MAX_PATHNAME 32
#define MAX_FULLNAME (MAX_DBNAME + MAX_PATHNAME + 2)

/* StrNChr is implemented in terms of StrChr which ensures we get WChar support for free */
static Char *StrNChr(const Char *p, WChar c)
{
	Char *q;
	for (p = q = StrChr(p, c); p; q = (Char *) p, p = StrChr(p+1, c)) {}	
	return q;
}

static Err VFSFileDBInfoFromPath(
		UInt16 volRefNum, const Char *pathNameP,
		Char *nameP, UInt16 *attributesP, UInt16 *versionP, UInt32 *crDateP, UInt32 *modDateP,
		UInt32 *bckUpDateP, UInt32 *modNumP, MemHandle *appInfoHP, MemHandle *sortInfoHP,
		UInt32 *typeP, UInt32 *creatorP, UInt16 *numRecordsP)
{
	Err err;
	FileRef fileRef;
	
	err = VFSFileOpen(volRefNum, pathNameP, vfsModeRead, &fileRef);
	if (errNone == err) {
		err = VFSFileDBInfo(fileRef, nameP, attributesP, versionP, crDateP, modDateP, bckUpDateP,
				    modNumP, appInfoHP, sortInfoHP, typeP, creatorP, numRecordsP);
		(void) VFSFileClose(fileRef);
	}
	
	return err;
}

static Err VFSFileVPrintF(FileRef fileRef, const Char *fmt, va_list args)
{
	Char line[80];
	Int16 count;
	
	count = StrVPrintF(line, fmt, args);
	ErrNonFatalDisplayIf(count > sizeof(line), "internal string overflow");
	
	return VFSFileWrite(fileRef, count, line, NULL);
}

/*static*/ Err VFSFilePrintF(FileRef fileRef, const Char *fmt, ...)
{
	Err err;
	va_list args;
	
	va_start(args, fmt);
	err = VFSFileVPrintF(fileRef, fmt, args);
	va_end(args);
	return err;
}


static Err BakLogToFile(FileRef fileRef, UInt16 resID, ...)
{
	va_list args;
	MemHandle hdl;
	const Char *fmt;
	Err err;
	
	va_start(args, resID);
	hdl = DmGetResource('tSTR', resID);
	ErrNonFatalDisplayIf(0 == hdl, "Resource database incomplete/corrupt");
	fmt = (Char *) MemHandleLock(hdl);
	err = VFSFileVPrintF(fileRef, fmt, args);
	MemHandleUnlock(hdl);
	DmReleaseResource(hdl);
	va_end(args);
	
	return err;
}

static const Char *BakGetTime(void)
{
	DateTimeType dateTime;
	static Char time[timeStringLength+1];
	
	TimSecondsToDateTime(TimGetSeconds(), &dateTime);
	/* TODO: the format type should be fetched from the preferences */
	TimeToAscii(dateTime.hour, dateTime.minute, tfColonAMPM, time);
	return time;
}

static const Char *BakGetDate(void)
{
	DateTimeType dateTime;
	static Char date[longDateStrLength+1];
	
	TimSecondsToDateTime(TimGetSeconds(), &dateTime);
	/* TODO: the format type should be fetched from the preferences */
	DateToAscii(dateTime.month, dateTime.day, dateTime.year, dfDMYLong, date);
	return date;
}

static Boolean BakIsSpecialFilename(const Char *fname)
{
	/* TODO: would it be better to return false for .prc and .pdb and true for everything else */
	if (0 == StrCompareAscii(fname, "DATABASE.TMP") ||
	    0 == StrCompareAscii(fname, "BACKUP.LOG")) {
	    	return true;
	}
	
	return false;
}

static void BakStripIllegalCharacters(Char *s)
{
	Char illegal[] = ":/\\?", *i, *p;
	
	for (i = illegal; *i != '\0'; i++) {
		for (p = StrChr(s, *i); p; p = StrChr(p+1, *i)) {
			*p = '_';
		}
	}
}

Err BakCleanupBackupDirectory(UInt16 volRefNum, const Char *pathNameP, BakCallback *cb, FileRef logFile)
{
	Err err;
	FileRef dirRef;
	UInt32 dirIterator;
	FileInfoType info;
	MemHandle strHandle;
	Char *str, strName[MAX_FILENAME];
	Char fileName[MAX_FILENAME];
	
	/* get the deleting string */
	strHandle = DmGetResource('tSTR', kStringDeleting);
	ErrNonFatalDisplayIf(0 == strHandle, "kStringDeleting is missing");
		
	err = VFSFileOpen(volRefNum, pathNameP, vfsModeRead, &dirRef);
	if (errNone != err) {
		return err;
	}
	
	/* initialize the info structure */
	info.nameP = fileName;
	info.nameBufLen = MAX_FILENAME;	
	
	/* iterate over every file in the backup directory */
	for (dirIterator = vfsIteratorStart, err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info);
	     errNone == err && vfsIteratorStop != dirIterator;
	     err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info)) {
	     	Char name[MAX_DBNAME];
	     	Char pathName[MAX_FULLNAME];
   		LocalID dbID;

	     	/* guarantee no power off while restore is in progress */
	     	(void) EvtResetAutoOffTimer();

		if (BakIsSpecialFilename(fileName)) {
			continue;
		}

		/* generate the full path name for the .prc or .pdb file */
	     	StrCopy(pathName, pathNameP);
	     	StrNCat(pathName, "/", sizeof(pathName));
	     	StrNCat(pathName, fileName, sizeof(pathName));
	     
	        /* obtain the original name of the database */
	     	err = VFSFileDBInfoFromPath(volRefNum, pathName, name,
	     	                            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
	     	if (errNone != err) {
	     		ErrAlert(err);
	     		continue;
	     	}
	     	
	     	/* lookup the name on the Palm device and remove it from the backup
	     	 * it it does not exist
	     	 */
	     	dbID = DmFindDatabase(0, name);
		if (0 == dbID) {
			/* update the GUI */
			str = (Char *) MemHandleLock(strHandle);
			StrPrintF(strName, str, name);
			MemHandleUnlock(strHandle);
			cb(strName);
			
			(void) VFSFileDelete(volRefNum, pathName);
			(void) BakLogToFile(logFile, kStringRemovedFromCard, name);
			
			/* as soon as we have deleted a file from the directory any outstanding iterators
			 * become obsolete. we will therefore have to walk the entire directory again.
			 */
			dirIterator = vfsIteratorStart;
		}
	}
	
	DmReleaseResource(strHandle);
	
	VFSFileClose(dirRef);
	
	return err;
}

static Boolean BakCompareDatabase(UInt16 volRefNum, const Char *pathNameP, UInt16 cardNo, LocalID dbID)
{
	Err err;
	
	UInt16 palmAttributes, cardAttributes;
	UInt16 palmVersion, cardVersion;
	UInt32 palmModDate, cardModDate;
	UInt32 palmModNum, cardModNum;
	UInt32 palmType, cardType;
	UInt32 palmCreator, cardCreator;
	
	err = DmDatabaseInfo(cardNo, dbID, 0, &palmAttributes, &palmVersion,
	                     0, &palmModDate, 0, &palmModNum, 0, 0, &palmType, &palmCreator);
	ErrNonFatalDisplayIf(errNone != err, "DmDatabaseInfo() failed after previously succeeding");
		
	err = VFSFileDBInfoFromPath(volRefNum, pathNameP, 0, &cardAttributes, &cardVersion, 0,
				    &cardModDate, 0, &cardModNum, 0, 0, &cardType, &cardCreator, 0);
	
	if (errNone == err &&
	    palmAttributes == cardAttributes && palmVersion == cardVersion &&
	    palmModDate == cardModDate && palmModNum == cardModNum &&
	    palmType == cardType && palmCreator == cardCreator) {
	    	return true;
	}
	
	return false;
}		

static Err BakBackupDatabase(UInt16 volRefNum, const Char *pathNameP, UInt16 cardNo, LocalID dbID)
{
	Char tmpName[MAX_FULLNAME];
	Char *fileName;
	Err err;
		
	/* generate the supplementary filenames */
	StrCopy(tmpName, pathNameP);
	fileName = StrNChr(tmpName, '/') + 1;
	ErrNonFatalDisplayIf(0 == (fileName - 1), "Badly formed pathNameP");
	/* this name must be 8.3 otherwise bugs in the PalmOS FAT driver start making odd things happen */
	StrCopy(fileName, "DATABASE.TMP");
	fileName = StrNChr(pathNameP, '/') + 1;

	/* delete obsolete temporary file (if it exists ) */
	(void) VFSFileDelete(volRefNum, tmpName);
		
	/* store the new version of the database */
	err = VFSExportDatabaseToFile (volRefNum, tmpName, cardNo, dbID);
	ErrNonFatalDisplayIf(errNone != err, "VFSExportDatabaseToFile() failed");
	if (errNone != err) {
		return err;
	}

	/* delete the original backup copy (if it exists) */
	(void) VFSFileDelete(volRefNum, pathNameP);
		
	/* rename the temporary copy */
	err = VFSFileRename (volRefNum, tmpName, fileName);
	ErrNonFatalDisplayIf(errNone != err, "VFSFileRename() failed");
	
	return err;
}

Err BakBackupToVFS(UInt16 volRefNum, const Char *pathNameP, BakCallback *cb)
{
	Err err;
	MemHandle strHandle;
	FileRef logFile;
	const Char *str;
	DmSearchStateType stateInfo;
	UInt16 cardNo;
	LocalID dbID;

	/* ensure the backup path exists */
	ErrNonFatalDisplayIf(StrLen(pathNameP) >= MAX_PATHNAME, "pathNameP is too long");
	/* TODO: hard coded path */
	(void) VFSDirCreate(volRefNum, "/PALM");
	(void) VFSDirCreate(volRefNum, "/PALM/RFBackup");
	/* TODO: error checking (problems on emulator (when file exists vfsErrFileGeneric is returned) */

	/* open the log file */
	/* TODO: hard coded path */
	(void) VFSFileDelete(volRefNum, "/PALM/RFBackup/BACKUP.LOG");
	err = VFSFileOpen(volRefNum, "/PALM/RFBackup/BACKUP.LOG",
	                  vfsModeWrite | vfsModeCreate, &logFile);
	if (errNone != err) {
		ErrAlert(err);
	}
	(void) BakLogToFile(logFile, kStringBackupStartedAt, BakGetDate(), BakGetTime());
		
	/* clean up any obsolete files from the backup directory */
	strHandle = DmGetResource('tSTR', kStringCleaningUp);
	ErrNonFatalDisplayIf(0 == strHandle, "kStringCleaningUp is missing");
	str = (Char *) MemHandleLock(strHandle);
	cb(str);
	MemHandleUnlock(strHandle);
	DmReleaseResource(strHandle);
	
	err = BakCleanupBackupDirectory(volRefNum, pathNameP, cb, logFile);
	if (errNone != err) {
		ErrAlert(err);
	}
	
	/* iterate over every database in the system */
	for (err = DmGetNextDatabaseByTypeCreator(true,  &stateInfo, 0, 0, false, &cardNo, &dbID);
	     errNone == err && 0 != dbID;
	     err = DmGetNextDatabaseByTypeCreator(false, &stateInfo, 0, 0, false, &cardNo, &dbID)) {
	     	Char name[MAX_DBNAME];
	     	Char pathName[MAX_FULLNAME];
	     	UInt16 attributes, len;
	     	
	     	/* guarantee no power off while backup is in progress */
	     	(void) EvtResetAutoOffTimer();
	     	
	     	/* get database attributes so we can generate the proper filename */
	     	err = DmDatabaseInfo(cardNo, dbID, name, &attributes, 0, 0, 0, 0, 0, 0, 0, 0, 0);

		/* skip this database on error of if the database is held in ROM */
	     	if (errNone != err || 0 != (attributes & dmHdrAttrReadOnly)) {
	     		continue;
	     	}

	     	/* update the GUI */
	     	cb(name);
	     	
	     	/* generate a complete filename for the database */
	     	BakStripIllegalCharacters(name);
	     	len = StrPrintF(pathName, "%s/%s%s", pathNameP, name,
	     	                0 == (attributes & dmHdrAttrResDB) ? ".pdb" : ".prc");
	     	ErrNonFatalDisplayIf(len > sizeof(pathName), "string overflow");
	     		     	
	     	/* perform the actual backup (if required) */
	     	if (!BakCompareDatabase(volRefNum, pathName, cardNo, dbID)) {
	     		err = BakBackupDatabase(volRefNum, pathName, cardNo, dbID);
	     		if (errNone != err) {
	     			ErrAlert(err);
	     			(void) BakLogToFile(logFile, kStringBackupFailed, name);
			} else {
				(void) BakLogToFile(logFile, kStringBackedUp, name);
	     		}
	     	} else {
	     		BakLogToFile(logFile, kStringUpToDate, name);
	     	}
	}
			
	/* ... and we're done with the log file */		
	(void) BakLogToFile(logFile, kStringBackupCompletedAt, BakGetDate(), BakGetTime());
	VFSFileClose(logFile);
	
	return dmErrCantFind == err ? errNone : err;
}

Err BakRestoreFromVFS(UInt16 volRefNum, const Char *pathNameP, BakCallback *cb)
{
	Err err;
	FileRef dirRef;
	UInt32 dirIterator;
	FileInfoType info;
	Char fileName[MAX_FILENAME];
		
	err = VFSFileOpen(volRefNum, pathNameP, vfsModeRead, &dirRef);
	if (errNone != err) {
		return err;
	}
	
	/* initialize the info structure */
	info.nameP = fileName;
	info.nameBufLen = MAX_FILENAME;	
	
	/* iterate over every file in the backup directory */
	for (dirIterator = vfsIteratorStart, err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info);
	     errNone == err && vfsIteratorStop != dirIterator;
	     err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info)) {
	     	Char name[MAX_DBNAME];
	     	UInt32 creator;
	     	Char pathName[MAX_FULLNAME];
   		UInt16 cardNo;
		LocalID dbID;

	     	/* guarantee no power off while restore is in progress */
	     	(void) EvtResetAutoOffTimer();

		if (BakIsSpecialFilename(fileName)) {
			continue;
		}

		/* generate the full path name for the .prc or .pdb file */
	     	StrCopy(pathName, pathNameP);
	     	StrNCat(pathName, "/", sizeof(pathName));
	     	StrNCat(pathName, fileName, sizeof(pathName));
	     
	        /* obtain the original name of the database */
	     	err = VFSFileDBInfoFromPath(volRefNum, pathName, name,
	     	                            0, 0, 0, 0, 0, 0, 0, 0, 0, &creator, 0);
	     	if (errNone != err) {
	     		ErrAlert(err);
	     		continue;
	     	}
	     	
	     	/* skip restore of RFBackup itself */
	     	if (creator == CREATOR_CODE) {
	     		continue;
	     	}
	     	
	     	/* update the gui */
	     	cb(name);
	     	
		/* delete the original database */
		dbID = DmFindDatabase(0, name);
		if (0 != dbID) {
			err = DmDeleteDatabase(0, DmFindDatabase(0, name));
			
			if (dmErrDatabaseProtected == err) {
				/* TODO: experiment with DmDatabaseProtect (allows restore of Net Prefs) 
				 * for now we'll just skip this one
				 */
				continue;
			}
			
			if (errNone != err) {
				ErrAlert(err);
				continue;
			}
		}
	     	
	     	/* finally restore the database from the backup */
	     	err = VFSImportDatabaseFromFile(volRefNum, pathName, &cardNo, &dbID);
	      	if (errNone != err) {
	     		break;
		}	
	} 
	
	VFSFileClose(dirRef);
	
	return err;
}